% Bungy Jumping with rigid object
% Andy Ruina,  April 11, 2019
% Uses rhs function MYBUNGY, below
% Uses rotate function ROTATEBYTHETA, below
% Decently commented (unlike the simpler version)

%Simulate and animate rigid bungy jumper
clear all,  close all

%Make Fonts for plots bigger. Good to put in STARTUP.M
set(0,'DefaultAxesFontSize', 30)

%PARAMETERS
p.L0  = 10; p.k   = 1; p.ell= 20; p.w= 10; p.g= 1; %Problem parameters
p.m = 10;   p.IG  = p.m*(p.ell^2 + p.w^2)/12;
%%%%%%%%%%%

%INITIAL CONDITIONS
rG0 = [20; 5]; theta0 = pi/2;      
vG0 = [0; 0]; omega0 = 0;
z0 = [rG0; theta0; vG0; omega0];
%%%%%%%%%%%%%%%%%%%%

%SETUP ODE SOLUTION
tarray            = linspace(0,200,1000);   % output times
f                 = @(t,z)  mybungy(z,p);  % rhs function
small             = 1e-8;                  % accuracy
options           = odeset('RelTol', small, 'AbsTol', small);
%%%%%%%%%%%%%%%%%%%%

%SOLVE ODES HERE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[tarray, zarray]  = ode45(f, tarray, z0, options);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Unpack solution
xG  = zarray(:,1);    yG = zarray(:,2);   theta = zarray(:,3); 
vxG = zarray(:,4);   vyG = zarray(:,5);   omega = zarray(:,6);
rG  = [xG,yG];

%ENERGY CHECK%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
KE = p.m*(vxG.^2 + vyG.^2)/2 + p.IG* omega.^2/2; %kinetic energy
rGA0    = [ p.ell/2; p.w/2; 0 ];     % A rel. to G, before rotat.

rGA = zeros(3,length(tarray));  %Next several lines are about spring stretch
for i = 1: length(tarray)
  rGA(:,i)     = rotatebytheta(rGA0,theta(i)); % function defined below 
end
rA      = rG + rGA(1:2,:)';   % Corner of box A rel. to origin 0, one row for each time
L       = sqrt(rA(:,1).^2 + rA(:,2).^2);   % length of spring after stretch 
PE = p.m * p.g * yG +  p.k* (L - p.L0).^2/2;
E  = KE + PE;
fprintf('\n******\n')
fprintf('So callled ''Accuracy'', the variable called ''small'' = %3.2e\n', small)
fprintf('ENERGY CHECK: Energy variation scaled by size of energy:\n')
fprintf('(max(E) - min(E))/max(sqrt(E.^2 + KE.^2 + PE.^2))  = %3.2e\n', ...
                      (max(E) - min(E))/max(sqrt(E.^2 + KE.^2 + PE.^2)/3))
fprintf('  (If ''small'' is small, so should be the energy variation.)\n\n')
%End of Energy Check %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%DRAW A BOX, This is the reference box, before motion
box0 = [0 0     %unit box
        1 0
        1 1
        0 1
        0 0];
% Stretch to size and translate box to put middle at origin
% Make it into two rows instead of two columns.
box0 =  [(box0(:,1) -.5)*p.ell, (box0(:,2) -.5)*p.w]';

%Add a row of zeros for the rotation to work
box0 = [box0;zeros(1,5)]; %first row is x, second is y, third is 0.
% END OF REFERENCE BOX DRAWING           


%ANIMATION LOOP
for i = 2:length(tarray)
    
%Rotate box and translate it 
box = rotatebytheta(box0, theta(i)) + [xG(i);yG(i);0];

spring = [ 0  box(1,3)
           0  box(2,3)];  %from origin to corner of box

%PLOT BOX AND SPRING AND POINTS
plot(box(1,3),box(2,3),'k.', 'MarkerSize', 20); hold on   % Dot at corner
patch(box(1,:), box(2,:),'r');   %PATCH makes a solid polygon
plot(spring(1,:),spring(2,:),'b','LineWidth', 3); % Spring is a blue line
plot(xG(i),yG(i),'.k', 'MarkerSize',20);        % Center of Box dot
plot(0,0,'k.', 'MarkerSize', 50); hold on     % Dot at the origin, 
hold off

%SET PICTURE FRAME BIG ENOUGH TO HOLD THE PICTURE
axis equal  % Has to be typed before limits are set
marg      = sqrt(p.ell^2 + p.w^2)/2;  %leave room for box in picture
ax_limits = ...  %Set max and min for x and y for plot size
    [min(xG)-marg, max(xG)+marg, min(yG)-marg, max([yG;0])+marg];
axis(ax_limits);  % make picture big enough to hold whole animation
%End of picture frame sizing

grid on;   xlabel('X'); ylabel('Y'); title('Animation of bungy')

hold off %try deleting "hold off" for stroboscopic effect

pause(.01)  % .01s is as short as Pause can do.
    end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%END OF ANIMATION



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% RHS function
function zdot = mybungy(z,p)
rG   = [z(1:2);0]; theta = z(3); % Unpack the state variables
vG   = [z(4:5);0]; omega = z(6); % Make all vectors 3D, for cross products
jhat = [ 0; 1; 0];               % needed for gravity force

rGA0    = [ p.ell/2; p.w/2; 0 ];     % A rel. to G, before rotat.
rGA     = rotatebytheta(rGA0,theta); % function defined below 
rA      = rG + rGA;   % Corner of box A rel. to origin 0.
L       = norm(rA);   % length of spring after stretch
lambdaA = rA/L;       % Unit vector from O towards A

Tspring =   p.k*(L - p.L0);     % (spring constant)  * stretch
Fspring = - Tspring * lambdaA;  % "magnitude" * direction

rGdot    = vG(1:2);     % throw away z component
thetadot = omega;
vGdot    = (Fspring - p.m * p.g * jhat)/p.m;  % LMB
vGdot    = vGdot(1:2);  % throw away z component
omegadot = cross(rGA,Fspring) / p.IG;   % AMB 
omegadot = omegadot(3);

zdot = [rGdot; thetadot;  vGdot;  omegadot];
end  %END of RHS FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%
function p = rotatebytheta(p0,theta)
% Rotate points by theta in xy plane
% p0 and p are both 3 X 1 vectors, third component is zero.
c = cos(theta); s = sin(theta);
R = [c -s  0;    % Use x,y,z comps so we can do cross products
     s  c  0;
     0  0  1 ];  % 2D rotation, does nothing to 3rd dim.
p = R*p0;        % Rotation about z axis
end  %END OF ROTATION FUNCTION
%%%%%%%%%%%%%%%%%

